#include "fs-compat.hpp"
#include "options.hpp"
#include "search.hpp"

// catch 3 does not have catch.hpp anymore
#if __has_include(<catch.hpp>)
#    include <catch.hpp>
#else
#    include <catch_all.hpp>
#endif

#include <clocale>
#include <exception>
#include <iostream>
#include <string>
#include <vector>

SCENARIO("Searching ZIP files works")
{
    GIVEN("Our test zip file")
    {
        fs::path zipfile{"test.zip"};
        std::setlocale(LC_CTYPE, ""); // Needed for utf-8 support in libarchive.
        bool exception{false};

        REQUIRE(fs::exists(zipfile));

        SECTION("search() doesn't fail and returns the right lines")
        {
            std::vector<epubgrep::search::match> matches;
            epubgrep::search::settings opts;
            opts.raw = true;

            WHEN("We search for ‘📙+\\w?’ using extended regular expressions")
            {
                try
                {
                    opts.regex = epubgrep::options::regex_kind::extended;
                    matches = epubgrep::search::search(zipfile, "📙+\\w?", opts);
                }
                catch (const std::exception &e)
                {
                    std::cerr << "EXCEPTION: " << e.what() << '\n';
                    exception = true;
                }

                THEN("No exception is thrown")
                AND_THEN("It returns the match correctly")
                {
                    REQUIRE_FALSE(exception);
                    REQUIRE(matches.at(0).filepath_inside == "test folder/😊");
                    REQUIRE(matches.at(0).text == "📙");
                }
            }

            WHEN("We search for ‘📗’ with context = 1")
            {
                try
                {
                    opts.context = 1;
                    matches = epubgrep::search::search(zipfile, "📗", opts);
                }
                catch (const std::exception &e)
                {
                    std::cerr << "EXCEPTION: " << e.what() << '\n';
                    exception = true;
                }

                THEN("No exception is thrown")
                AND_THEN("It returns the match correctly")
                {
                    REQUIRE_FALSE(exception);
                    REQUIRE(matches.at(0).filepath_inside == "test folder/😊");
                    REQUIRE(matches.at(0).text == "📗");
                    REQUIRE(matches.at(0).context.first == "📖\n\n📘");
                    REQUIRE(matches.at(0).context.second == "📙");
                }
            }

            WHEN("We search for ‘[ \\n]’ with context = 1.")
            {
                try
                {
                    opts.context = 1;
                    opts.regex = epubgrep::options::regex_kind::perl;
                    matches = epubgrep::search::search(zipfile, R"([ \n])",
                                                       opts);
                }
                catch (const std::exception &e)
                {
                    std::cerr << "EXCEPTION: " << e.what() << '\n';
                    exception = true;
                }

                THEN("No exception is thrown")
                AND_THEN("It returns the match correctly")
                {
                    // I looked at this a week or so after I've written it, and
                    // I have come to the realization that this is a tiny bit
                    // more complicated than strictly required. 😄
                    // TODO: Rewrite test.zip and tests to be better
                    // understandable.
                    REQUIRE_FALSE(exception);
                    REQUIRE(matches.at(1).filepath_inside
                            == "test folder/test file");
                    REQUIRE(matches.at(1).text == " ");
                    REQUIRE(matches.at(1).context.first == "don't");
                    REQUIRE(matches.at(1).context.second == "want to");
                    REQUIRE(matches.at(10).filepath_inside == "test folder/😊");
                    REQUIRE(matches.at(10).text == "\n");
                    REQUIRE(matches.at(10).context.first == "📖");
                    REQUIRE(matches.at(10).context.second == "\n📘📗📙");
                    REQUIRE(matches.at(12).filepath_inside == "test folder/😊");
                    REQUIRE(matches.at(12).text == "\n");
                    REQUIRE(matches.at(12).context.first == "📘📗📙");
                    REQUIRE(matches.at(12).context.second.empty());
                }
            }

            WHEN(R"(We search for ‘work\s[\w]+\.\W[\w']+\Wstay’ )"
                 R"(using Perl regular expressions. context = 1.)")
            {
                try
                {
                    opts.context = 1;
                    opts.regex = epubgrep::options::regex_kind::perl;
                    matches = epubgrep::search::search(
                        zipfile, R"(work\s[\w]+\.\W[\w']+\Wstay)", opts);
                }
                catch (const std::exception &e)
                {
                    std::cerr << "EXCEPTION: " << e.what() << '\n';
                    exception = true;
                }

                THEN("No exception is thrown")
                AND_THEN("It returns the match correctly")
                {
                    REQUIRE_FALSE(exception);
                    REQUIRE(matches.at(0).filepath_inside
                            == "test folder/test file");
                    REQUIRE(matches.at(0).text == "work today.\nI'm stay");
                    REQUIRE(matches.at(0).context.first == "to ");
                    REQUIRE(matches.at(0).context.second == "ing in");
                }
            }
        }
    }
}
